##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import re
import os

class G:
    abi_h_lines = []
    abi_datatypes = {}
    abi_ops = {}
    op_mask = 0x1f;
    datatype_mask = 0x1ff;

class RE:
    m = None
    def match(pat, str, flags=0):
        RE.m = re.match(pat, str, flags)
        return RE.m
    def search(pat, str, flags=0):
        RE.m = re.search(pat, str, flags)
        return RE.m

def main():
    load_mpi_abi_h("src/binding/abi/mpi_abi.h")
    dump_mpi_abi_internal_h("src/binding/abi/mpi_abi_internal.h")
    dump_romio_abi_internal_h("src/mpi/romio/include/romio_abi_internal.h")
    dump_mpi_abi_util_c("src/binding/abi/mpi_abi_util.c")

def load_mpi_abi_h(mpi_abi_h):
    with open(mpi_abi_h, "r") as In:
        G.abi_h_lines = In.readlines()

def dump_mpi_abi_internal_h(mpi_abi_internal_h):
    define_constants = {}
    def gen_mpi_abi_internal_h(out):
        re_Handle = r'\bMPI_(Comm|Datatype|Errhandler|Group|Info|Message|Op|Request|Session|Win|File|KEYVAL_INVALID|TAG_UB|IO|HOST|WTIME_IS_GLOBAL|APPNUM|LASTUSEDCODE|UNIVERSE_SIZE|WIN_BASE|WIN_DISP_UNIT|WIN_SIZE|WIN_CREATE_FLAVOR|WIN_MODEL)\b'
        for line in G.abi_h_lines:
            if RE.search(r'MPI_ABI_H_INCLUDED', line):
                # skip the include guard, harmless
                pass
            elif RE.match(r'\s*int MPI_(SOURCE|TAG|ERROR);', line):
                # no need to rename status fields
                out.append(line.rstrip())
            elif RE.match(r'\s*(int|MPI_Fint) .*(reserved|internal)\[(\d+)\];', line):
                # internal fields in MPI_Status/MPI_F08_status
                T = RE.m.group(1)
                n = int(RE.m.group(3))
                out.append("    %s count_lo;" % T)
                out.append("    %s count_hi_and_cancelled;" % T)
                out.append("    %s reserved[%d];" % (T, n - 2))
            elif RE.match(r'#define\s+(MPI_\w+)\s+\(?\((MPI_\w+)\)\s*(0x\w+)\)?', line):
                (name, T, val) = RE.m.group(1,2,3)
                if T == "MPI_Datatype":
                    idx = int(val, 0) & G.datatype_mask
                    G.abi_datatypes[idx] = name
                elif T == "MPI_Op":
                    idx = int(val, 0) & G.op_mask
                    G.abi_ops[idx] = name
                # replace param prefix
                out.append(re.sub(r'\bMPI_', 'ABI_', line.rstrip()))
            elif RE.match(r'#define MPI_(LONG_LONG|C_COMPLEX)', line):
                # datatype aliases
                out.append(re.sub(r'\bMPI_', 'ABI_', line.rstrip()))
            elif RE.match(r'^typedef struct \w+\s*\*\s*(MPI_T_\w+);', line):
                # rename to internal mpi_t struct type
                T = RE.m.group(1)
                out.append("typedef struct %s_s *%s;" % (re.sub(r'MPI_', 'MPIR_', T), T))
            elif RE.match(r'^(int|double|MPI_\w+) (P?MPI\w+)\((.*)\);', line):
                # prototypes, rename param prefix, add MPICH_API_PUBLIC
                (T, name, param) = RE.m.group(1,2,3)
                T=re.sub(re_Handle, r'ABI_\1', T)
                param=re.sub(re_Handle, r'ABI_\1', param)
                out.append("%s %s(%s) MPICH_API_PUBLIC;" % (T, name, param))
            elif RE.match(r'^extern\s+(\w+\s*\**)\s+(MPI_\w+);', line):
                (T, name) = RE.m.group(1,2)
                out.append("extern %s %s MPICH_API_PUBLIC;" % (T, name))
            elif RE.match(r'\s*(MPI_THREAD_\w+)\s*=\s*(\d+)', line):
                # These constants need to be converted from enum-constants into macro-constants
                define_constants[RE.m.group(1)] = RE.m.group(2)
                pass
            else:
                # replace param prefix
                out.append(re.sub(re_Handle, r'ABI_\1', line.rstrip()))

    # ----
    output_lines = []
    gen_mpi_abi_internal_h(output_lines)

    print(" --> [%s]" % mpi_abi_internal_h)
    with open(mpi_abi_internal_h, "w") as Out:
        dump_copyright(Out)
        print("#ifndef MPI_ABI_INTERNAL_H_INCLUDED", file=Out)
        print("#define MPI_ABI_INTERNAL_H_INCLUDED", file=Out)
        print("", file=Out)

        for k in define_constants:
            print("#define %s %s" % (k, define_constants[k]), file=Out)
        print("", file=Out)
        for line in output_lines:
            print(line, file=Out)
        print("", file=Out)
        
        print("#endif /* MPI_ABI_INTERNAL_H_INCLUDED */", file=Out)

def dump_romio_abi_internal_h(romio_abi_internal_h):
    def gen_romio_abi_internal_h(out):
        for line in G.abi_h_lines:
            if RE.search(r'MPI_ABI_H_INCLUDED', line):
                # skip the include guard, harmless
                pass
            elif RE.match(r'typedef struct.*\bMPI_File;\s*$', line):
                out.append("typedef struct ADIOI_FileD *MPI_File;")
            elif RE.match(r'(int|double|MPI_\w+) (P?MPI\w+)\((.*)\);', line):
                # prototypes, rename param prefix, add MPICH_API_PUBLIC
                (T, name, param) = RE.m.group(1,2,3)
                if RE.match(r'P?MPI_(File_\w+|Register_datarep\w*)', name):
                    out.append("%s %s(%s) ROMIO_API_PUBLIC;" % (T, name, param))
                else:
                    out.append("%s %s(%s);" % (T, name, param))
            else:
                # replace param prefix
                out.append(line.rstrip())

    def add_romio_visibility(out):
        out.append("#if defined(HAVE_VISIBILITY)")
        out.append("#define ROMIO_API_PUBLIC __attribute__((visibility (\"default\")))")
        out.append("#else")
        out.append("#define ROMIO_API_PUBLIC")
        out.append("#endif")
        out.append("")

    def add_mpix_grequest(out):
        out.append("#define MPIO_Request MPI_Request")
        out.append("#define MPIO_USES_MPI_REQUEST")
        out.append("#define MPIO_Wait MPI_Wait")
        out.append("#define MPIO_Test MPI_Test")
        out.append("#define PMPIO_Wait PMPI_Wait")
        out.append("#define PMPIO_Test PMPI_Test")
        out.append("#define MPIO_REQUEST_DEFINED")
        out.append("")
        # they are not in the header, but they are in c_binding_abi.c
        out.append("typedef int MPIX_Grequest_class;")
        out.append("typedef int (MPIX_Grequest_poll_function)(void *, MPI_Status *);")
        out.append("typedef int (MPIX_Grequest_wait_function)(int, void **, double, MPI_Status *);")
        out.append("int MPIX_Grequest_start(MPI_Grequest_query_function *query_fn, MPI_Grequest_free_function *free_fn, MPI_Grequest_cancel_function *cancel_fn, MPIX_Grequest_poll_function *poll_fn, MPIX_Grequest_wait_function *wait_fn, void *extra_state, MPI_Request *request);")
        out.append("int MPIX_Grequest_class_create(MPI_Grequest_query_function *query_fn, MPI_Grequest_free_function *free_fn, MPI_Grequest_cancel_function *cancel_fn, MPIX_Grequest_poll_function *poll_fn, MPIX_Grequest_wait_function *wait_fn, MPIX_Grequest_class *greq_class);")
        out.append("int MPIX_Grequest_class_allocate(MPIX_Grequest_class greq_class, void *extra_state, MPI_Request *request);")
        out.append("int PMPIX_Grequest_start(MPI_Grequest_query_function *query_fn, MPI_Grequest_free_function *free_fn, MPI_Grequest_cancel_function *cancel_fn, MPIX_Grequest_poll_function *poll_fn, MPIX_Grequest_wait_function *wait_fn, void *extra_state, MPI_Request *request);")
        out.append("int PMPIX_Grequest_class_create(MPI_Grequest_query_function *query_fn, MPI_Grequest_free_function *free_fn, MPI_Grequest_cancel_function *cancel_fn, MPIX_Grequest_poll_function *poll_fn, MPIX_Grequest_wait_function *wait_fn, MPIX_Grequest_class *greq_class);")
        out.append("int PMPIX_Grequest_class_allocate(MPIX_Grequest_class greq_class, void *extra_state, MPI_Request *request);")
        out.append("")

    def add_mpix_iov(out):
        out.append("typedef struct MPIX_Iov {")
        out.append("    void *iov_base;")
        out.append("    MPI_Aint iov_len;")
        out.append("} MPIX_Iov;")
        out.append("")
        out.append("int MPIX_Type_iov_len(MPI_Datatype datatype, MPI_Count max_iov_bytes, MPI_Count *iov_len, MPI_Count *actual_iov_bytes);")
        out.append("int MPIX_Type_iov(MPI_Datatype datatype, MPI_Count iov_offset, MPIX_Iov *iov, MPI_Count max_iov_len, MPI_Count *actual_iov_len);")
        out.append("")

    def add_other(out):
        out.append("#define MPICH")  # ref. mpi-io/mpich_fileutil.c
        out.append("#define MPIR_ERRORS_THROW_EXCEPTIONS (MPI_Errhandler)0")

    # ----
    output_lines = []
    add_romio_visibility(output_lines)
    gen_romio_abi_internal_h(output_lines)
    add_mpix_grequest(output_lines)
    add_mpix_iov(output_lines)
    add_other(output_lines)

    print(" --> [%s]" % romio_abi_internal_h)
    with open(romio_abi_internal_h, "w") as Out:
        dump_copyright(Out)
        print("#ifndef ROMIO_ABI_INTERNAL_H_INCLUDED", file=Out)
        print("#define ROMIO_ABI_INTERNAL_H_INCLUDED", file=Out)
        print("", file=Out)

        for line in output_lines:
            print(line, file=Out)
        print("", file=Out)
        
        print("#endif /* ROMIO_ABI_INTERNAL_H_INCLUDED */", file=Out)

def dump_mpi_abi_util_c(mpi_abi_util_c):
    def dump_init_once(Out):
        print("    static int initialized = 0;", file=Out)
        print("    if (initialized) {", file=Out)
        print("        return;", file=Out)
        print("    }", file=Out)
        print("    initialized = 1;", file=Out)
        print("", file=Out)
    # ----
    print(" --> [%s]" % mpi_abi_util_c)
    with open(mpi_abi_util_c, "w") as Out:
        dump_copyright(Out)
        print("#include \"mpiimpl.h\"", file=Out)
        print("#include \"mpi_abi_util.h\"", file=Out)
        print("", file=Out)
        print("MPI_Datatype abi_datatype_builtins[ABI_MAX_DATATYPE_BUILTINS];", file=Out)
        print("MPI_Op abi_op_builtins[ABI_MAX_OP_BUILTINS];", file=Out)
        print("", file=Out)
        print("void ABI_init_builtins(void)", file=Out)
        print("{", file=Out)
        dump_init_once(Out)
        for idx in G.abi_datatypes.keys():
            print("    abi_datatype_builtins[%d] = %s;" % (idx, G.abi_datatypes[idx]), file=Out)
        print("", file=Out)
        for idx in G.abi_ops.keys():
            print("    abi_op_builtins[%d] = %s;" % (idx, G.abi_ops[idx]), file=Out)
        print("}", file=Out)
    num_datatypes = len(G.abi_datatypes)
    G.abi_h_lines.append("#define ABI_MAX_DATATYPE_BUILTINS %d" % num_datatypes)

def dump_copyright(Out):
    print("/*", file=Out)
    print(" * Copyright (C) by Argonne National Laboratory", file=Out)
    print(" *     See COPYRIGHT in top-level directory", file=Out)
    print(" */", file=Out)
    print("/* This file is automatically generated. DO NOT EDIT. */", file=Out)
    print("", file=Out)

#----------------------------------------
if __name__ == "__main__":
    main()



##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

from local_python import MPI_API_Global as G
from local_python.mpi_api import *
from local_python.binding_common import *
from local_python.binding_f77 import *
from local_python import RE
import os

def main():
    # currently support -no-real128, -no-mpiio, -aint-is-int
    G.parse_cmdline()

    binding_dir = G.get_srcdir_path("src/binding")
    f77_dir = "src/binding/fortran/mpif_h"

    func_list = load_C_func_list(binding_dir, True) # suppress noise

    if "no-mpiio" in G.opts:
        # a few MPI_File_xxx functions are already in (MPI_File_xxx_errhandler)
        func_list = [f for f in func_list if not f['name'].startswith('MPI_File_')]
    else:
        # FIXME: until romio interface is generated
        func_list.extend(get_mpiio_func_list())
    func_list.extend(get_f77_dummy_func_list())
    func_list.append(G.FUNCS['mpi_f_sync_reg'])

    # preprocess
    for func in func_list:
        check_func_directives(func)
    func_list = [f for f in func_list if '_skip_fortran' not in f]

    # fortran_binding.c
    def has_cptr(func):
        for p in func['parameters']:
            if p['kind'] == 'C_BUFFER':
                return True
        return False

    G.out = []
    G.profile_out = []
    for func in func_list:
        G.out.append("")
        dump_f77_c_func(func)
        if has_cptr(func):
            dump_f77_c_func(func, True)

    f = "%s/fortran_binding.c" % f77_dir
    dump_f77_c_file(f, G.out)

    f = "%s/fortran_profile.h" % f77_dir
    dump_f77_c_file(f, G.profile_out)

    # .in files has to be generated in the source tree
    if G.is_autogen():
        G.mpih_defines = {}
        load_mpi_h_in("src/include/mpi.h.in")
        load_mpi_h_in("src/mpi/romio/include/mpio.h.in")
        f = "%s/mpif.h.in" % f77_dir
        dump_mpif_h(f)

# ---------------------------------------------------------
if __name__ == "__main__":
    main()




#! /usr/bin/env python3
##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

import sys
import argparse

sys.path.append('maint/')
import yutils

sys.path.append('src/backend/')
import gencomm


########################################################################################
##### Global variables
########################################################################################

num_paren_open = 0
builtin_types = [ "_Bool", "char", "wchar_t", "int8_t", "int16_t", \
                  "int32_t", "int64_t", "float", "double", "long double", "c_complex", "c_double_complex", "c_long_double_complex" ]
blklens = [ "1", "2", "3", "4", "5", "6", "7", "8", "generic" ]
builtin_maps = { }


########################################################################################
##### Type-specific functions
########################################################################################

## hvector routines
def hvector_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "intptr_t count%d = %s->u.hvector.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t blocklength%d ATTRIBUTE((unused)) = %s->u.hvector.blocklength;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.hvector.stride;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def hvector(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (intptr_t j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    if (blklen == "generic"):
        yutils.display(OUTFILE, "for (intptr_t k%d = 0; k%d < blocklength%d; k%d++) {\n" % (suffix, suffix, suffix, suffix))
    else:
        yutils.display(OUTFILE, "for (intptr_t k%d = 0; k%d < %s; k%d++) {\n" % (suffix, suffix, blklen, suffix))
    global s
    if (last != 1):
        s += " + j%d * stride%d + k%d * extent%d" % (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + j%d * stride%d + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## blkhindx routines
def blkhindx_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "intptr_t count%d = %s->u.blkhindx.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t blocklength%d ATTRIBUTE((unused)) = %s->u.blkhindx.blocklength;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t *restrict array_of_displs%d = %s->u.blkhindx.array_of_displs;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def blkhindx(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (intptr_t j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    if (blklen == "generic"):
        yutils.display(OUTFILE, "for (intptr_t k%d = 0; k%d < blocklength%d; k%d++) {\n" % (suffix, suffix, suffix, suffix))
    else:
        yutils.display(OUTFILE, "for (intptr_t k%d = 0; k%d < %s; k%d++) {\n" % (suffix, suffix, blklen, suffix))
    global s
    if (last != 1):
        s += " + array_of_displs%d[j%d] + k%d * extent%d" % \
             (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + array_of_displs%d[j%d] + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## hindexed routines
def hindexed_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "intptr_t count%d = %s->u.hindexed.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t *restrict array_of_blocklengths%d = %s->u.hindexed.array_of_blocklengths;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t *restrict array_of_displs%d = %s->u.hindexed.array_of_displs;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def hindexed(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 2
    yutils.display(OUTFILE, "for (intptr_t j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    yutils.display(OUTFILE, "for (intptr_t k%d = 0; k%d < array_of_blocklengths%d[j%d]; k%d++) {\n" % \
            (suffix, suffix, suffix, suffix, suffix))
    global s
    if (last != 1):
        s += " + array_of_displs%d[j%d] + k%d * extent%d" % \
             (suffix, suffix, suffix, suffix + 1)
    else:
        s += " + array_of_displs%d[j%d] + k%d * sizeof(%s)" % (suffix, suffix, suffix, b)

## contig routines
def contig_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "intptr_t count%d = %s->u.contig.count;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "intptr_t stride%d = %s->u.contig.child->extent;\n" % (nesting, dtp))
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def contig(suffix, b, blklen, last):
    global num_paren_open
    num_paren_open += 1
    yutils.display(OUTFILE, "for (intptr_t j%d = 0; j%d < count%d; j%d++) {\n" % (suffix, suffix, suffix, suffix))
    global s
    s += " + j%d * stride%d" % (suffix, suffix)

# resized routines
def resized_decl(nesting, dtp, b):
    yutils.display(OUTFILE, "uintptr_t extent%d ATTRIBUTE((unused)) = %s->extent;\n" % (nesting, dtp))

def resized(suffix, b, blklen, last):
    pass


########################################################################################
##### Core kernels
########################################################################################
def generate_kernels(b, darray, blklen):
    global num_paren_open
    global s

    # individual blocklength optimization is only for
    # hvector and blkhindx
    if (len(darray) and darray[-1] != "hvector" and darray[-1] != "blkhindx" and blklen != "generic"):
        return

    for func in "pack","unpack":
        ##### figure out the function name to use
        s = "int yaksuri_seqi_%s_" % func
        for d in darray:
            s = s + "%s_" % d
        # hvector and hindexed get blklen-specific function names
        if (len(darray) and (darray[-1] == "hvector" or darray[-1] == "blkhindx")):
            s = s + "blklen_%s_" % blklen + b.replace(" ", "_")
        else:
            s = s + b.replace(" ", "_")
        yutils.display(OUTFILE, "%s(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type, yaksa_op_t op)\n" % s),
        yutils.display(OUTFILE, "{\n")


        ##### variable declarations
        # generic variables
        yutils.display(OUTFILE, "int rc = YAKSA_SUCCESS;\n");
        yutils.display(OUTFILE, "const char *restrict sbuf = (const char *) inbuf;\n");
        yutils.display(OUTFILE, "char *restrict dbuf = (char *) outbuf;\n");
        yutils.display(OUTFILE, "uintptr_t extent ATTRIBUTE((unused)) = type->extent;\n")
        yutils.display(OUTFILE, "\n");

        # variables specific to each nesting level
        s = "type"
        for x in range(len(darray)):
            getattr(sys.modules[__name__], "%s_decl" % darray[x])(x + 1, s, b)
            yutils.display(OUTFILE, "\n")
            s = s + "->u.%s.child" % darray[x]

        yutils.display(OUTFILE, "uintptr_t idx = 0;\n")

        ##### non-hvector and non-blkhindx
        yutils.display(OUTFILE, "switch (op) {\n")
        for op in gencomm.type_ops[b]:
            yutils.display(OUTFILE, "case YAKSA_OP__%s:\n" % op)
            yutils.display(OUTFILE, "{\n")

            yutils.display(OUTFILE, "for (intptr_t i = 0; i < count; i++) {\n")
            num_paren_open += 1
            s = "i * extent"
            for x in range(len(darray)):
                if (x != len(darray) - 1):
                    getattr(sys.modules[__name__], darray[x])(x + 1, b, "generic", 0)
                else:
                    getattr(sys.modules[__name__], darray[x])(x + 1, b, blklen, 1)

            type = b
            if (b == "c_complex"):
                b = "float _Complex"
            elif (b == "c_double_complex"):
                b = "double _Complex"
            elif (b == "c_long_double_complex"):
                b = "long double _Complex"

            if (func == "pack"):
                if ((b == "float" or b == "double" or b == "long double") and
                    (op == "MAX" or op == "MIN")):
                    yutils.display(OUTFILE, "YAKSURI_SEQI_OP_%s_FLOAT(%s, *((const %s *) (const void *) (sbuf + %s)), *((%s *) (void *) (dbuf + idx)));\n" % (op, b, b, s, b))
                else:
                    yutils.display(OUTFILE, "YAKSURI_SEQI_OP_%s(*((const %s *) (const void *) (sbuf + %s)), *((%s *) (void *) (dbuf + idx)));\n" % (op, b, s.replace(type,b), b))
            else:
                if ((b == "float" or b == "double" or b == "long double") and
                    (op == "MAX" or op == "MIN")):
                    yutils.display(OUTFILE, "YAKSURI_SEQI_OP_%s_FLOAT(%s, *((const %s *) (const void *) (sbuf + idx)), *((%s *) (void *) (dbuf + %s)));\n" % (op, b, b, b, s))
                else:
                    yutils.display(OUTFILE, "YAKSURI_SEQI_OP_%s(*((const %s *) (const void *) (sbuf + idx)), *((%s *) (void *) (dbuf + %s)));\n" % (op, b, b, s.replace(type,b)))

            yutils.display(OUTFILE, "idx += sizeof(%s);\n" % b)
            b = type
            for x in range(num_paren_open):
                yutils.display(OUTFILE, "}\n")
            num_paren_open = 0

            yutils.display(OUTFILE, "break;\n")
            yutils.display(OUTFILE, "}\n")
            yutils.display(OUTFILE, "\n")

        yutils.display(OUTFILE, "default:\n")
        yutils.display(OUTFILE, "    break;\n")
        yutils.display(OUTFILE, "}\n")

        yutils.display(OUTFILE, "\n");
        yutils.display(OUTFILE, "return rc;\n")
        yutils.display(OUTFILE, "}\n\n")


########################################################################################
##### main function
########################################################################################
if __name__ == '__main__':
    ##### parse user arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--pup-max-nesting', type=int, default=3, help='maximum nesting levels to generate')
    args = parser.parse_args()
    if (args.pup_max_nesting < 0):
        parser.print_help()
        print
        print("===> ERROR: pup-max-nesting must be positive")
        sys.exit(1)

    ##### generate the reduction kernels for contiguous types
    for b in builtin_types:
        filename = "src/backend/seq/pup/yaksuri_seqi_pup_%s.c" % b.replace(" ","_")
        yutils.copyright_c(filename)
        OUTFILE = open(filename, "a")
        yutils.display(OUTFILE, "#include <string.h>\n")
        yutils.display(OUTFILE, "#include <stdint.h>\n")
        yutils.display(OUTFILE, "#include <wchar.h>\n")
        yutils.display(OUTFILE, "#include \"yaksuri_seqi_pup.h\"\n")
        yutils.display(OUTFILE, "\n")

        emptylist = [ ]
        generate_kernels(b, emptylist, 0)

        OUTFILE.close()

    ##### generate the core pack/unpack kernels (single level)
    if args.pup_max_nesting > 0:
        for b in builtin_types:
            for d in gencomm.derived_types:
                filename = "src/backend/seq/pup/yaksuri_seqi_pup_%s_%s.c" % (d, b.replace(" ","_"))
                yutils.copyright_c(filename)
                OUTFILE = open(filename, "a")
                yutils.display(OUTFILE, "#include <string.h>\n")
                yutils.display(OUTFILE, "#include <stdint.h>\n")
                yutils.display(OUTFILE, "#include <wchar.h>\n")
                yutils.display(OUTFILE, "#include \"yaksuri_seqi_pup.h\"\n")
                yutils.display(OUTFILE, "\n")

                emptylist = [ ]
                emptylist.append(d)
                for blklen in blklens:
                    generate_kernels(b, emptylist, blklen)
                emptylist.pop()

                OUTFILE.close()

    ##### generate the core pack/unpack kernels (more than one level)
    if args.pup_max_nesting > 1:
        darraylist = [ ]
        yutils.generate_darrays(gencomm.derived_types, darraylist, args.pup_max_nesting - 2)
        for b in builtin_types:
            for d1 in gencomm.derived_types:
                for d2 in gencomm.derived_types:
                    filename = "src/backend/seq/pup/yaksuri_seqi_pup_%s_%s_%s.c" % (d1, d2, b.replace(" ","_"))
                    yutils.copyright_c(filename)
                    OUTFILE = open(filename, "a")
                    yutils.display(OUTFILE, "#include <string.h>\n")
                    yutils.display(OUTFILE, "#include <stdint.h>\n")
                    yutils.display(OUTFILE, "#include <wchar.h>\n")
                    yutils.display(OUTFILE, "#include \"yaksuri_seqi_pup.h\"\n")
                    yutils.display(OUTFILE, "\n")

                    for darray in darraylist:
                        darray.append(d1)
                        darray.append(d2)
                        for blklen in blklens:
                            generate_kernels(b, darray, blklen)
                        darray.pop()
                        darray.pop()

                    OUTFILE.close()

    ##### generate the core pack/unpack kernel declarations
    filename = "src/backend/seq/pup/yaksuri_seqi_pup.h"
    yutils.copyright_c(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "#ifndef YAKSURI_SEQI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_PUP_H_INCLUDED\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "#include <string.h>\n")
    yutils.display(OUTFILE, "#include <stdint.h>\n")
    yutils.display(OUTFILE, "#include \"yaksi.h\"\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_MAX(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = ((in) ^ (((in) ^ (out)) & -((in) < (out)))); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_MIN(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = ((out) ^ (((in) ^ (out)) & -((in) < (out)))); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_MAX_FLOAT(type,in,out) \\\n")
    yutils.display(OUTFILE, "    do { type x_[2] = { (in), (out) }; (out) = x_[(in) < (out)]; } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_MIN_FLOAT(type,in,out) \\\n")
    yutils.display(OUTFILE, "    do { type x_[2] = { (in), (out) }; (out) = x_[(in) > (out)]; } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_SUM(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) += (in); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_PROD(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) *= (in); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_LAND(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = ((out) && (in)); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_BAND(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) &= (in); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_LOR(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = ((out) || (in)); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_BOR(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) |= (in); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_LXOR(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = (!(out) != !(in)); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_BXOR(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) ^= (in); } while (0)\n")
    yutils.display(OUTFILE, "#define YAKSURI_SEQI_OP_REPLACE(in,out) \\\n")
    yutils.display(OUTFILE, "    do { (out) = (in); } while (0)\n")
    yutils.display(OUTFILE, "\n")

    darraylist = [ ]
    yutils.generate_darrays(gencomm.derived_types, darraylist, args.pup_max_nesting)
    for b in builtin_types:
        yutils.display(OUTFILE, "int yaksuri_seqi_pack_%s(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type, yaksa_op_t op);\n" % b.replace(" ", "_"))
        yutils.display(OUTFILE, "int yaksuri_seqi_unpack_%s(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type, yaksa_op_t op);\n" % b.replace(" ", "_"))
        for darray in darraylist:
            for blklen in blklens:
                # we don't need pup kernels for basic types
                if (len(darray) == 0):
                    continue

                # individual blocklength optimization is only for
                # hvector and blkhindx
                if (darray[-1] != "hvector" and darray[-1] != "blkhindx" \
                    and blklen != "generic"):
                    continue

                for func in "pack","unpack":
                    ##### figure out the function name to use
                    s = "int yaksuri_seqi_%s_" % func
                    for d in darray:
                        s = s + "%s_" % d
                    # hvector and hindexed get blklen-specific function names
                    if (len(darray) and darray[-1] != "hvector" and darray[-1] != "blkhindx"):
                        s = s + b.replace(" ", "_")
                    else:
                        s = s + "blklen_%s_" % blklen + b.replace(" ", "_")
                    yutils.display(OUTFILE, "%s" % s),
                    yutils.display(OUTFILE, "(const void *inbuf, void *outbuf, uintptr_t count, yaksi_type_s * type, yaksa_op_t op);\n")

    yutils.display(OUTFILE, "#endif  /* YAKSURI_SEQI_PUP_H_INCLUDED */\n")
    OUTFILE.close()

    ##### generate the pup makefile
    filename = "src/backend/seq/pup/Makefile.pup.mk"
    yutils.copyright_makefile(filename)
    OUTFILE = open(filename, "a")
    yutils.display(OUTFILE, "libyaksa_la_SOURCES += \\\n")
    for b in builtin_types:
        yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup_%s.c \\\n" % \
                       b.replace(" ","_"))
        if args.pup_max_nesting > 0:
            for d1 in gencomm.derived_types:
                yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup_%s_%s.c \\\n" % \
                               (d1, b.replace(" ","_")))
                if args.pup_max_nesting > 1:
                    for d2 in gencomm.derived_types:
                        yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup_%s_%s_%s.c \\\n" % \
                                       (d1, d2, b.replace(" ","_")))
    yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seq_pup.c\n")
    yutils.display(OUTFILE, "\n")
    yutils.display(OUTFILE, "noinst_HEADERS += \\\n")
    yutils.display(OUTFILE, "\tsrc/backend/seq/pup/yaksuri_seqi_pup.h\n")
    OUTFILE.close()

    ##### generate the switching logic to select pup functions
    gencomm.populate_pupfns(args.pup_max_nesting, "seq", blklens, builtin_types, builtin_maps)



##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##
#
# This script generates all MPI function prototypes for both
# MPI and PMPI prefixes.
#
# It serves as an example of how to use the python framework.
#

from local_python import MPI_API_Global as G
from local_python.mpi_api import *
from local_python.binding_c import *
from local_python import RE

def main():
    binding_dir = G.get_srcdir_path("src/binding")
    load_mpi_api(binding_dir + "/mpi_standard_api.txt")
    load_mpi_mapping(binding_dir + "/apis_mapping.txt")

    func_list = G.FUNCS.values()

    decls = []
    fndefs = []
    for func in func_list:
        if re.match(r'.*_(function)', func['name']):
            # these are function typedefs
            fndefs.append(get_declare_function(func, False))
            if function_has_POLY_parameters(func):
                fndefs.append(get_declare_function(func, True))
        else:
            decls.append(get_declare_function(func, False))
            if function_has_POLY_parameters(func):
                decls.append(get_declare_function(func, True))
    
    for l in fndefs:
        print("typedef " + l + ";")

    for l in decls:
        print(l + ";")

    for l in decls:
        # assume the first MPI_ is the function name
        print(re.sub(r'\b(MPI_\w+)\(', r'P\1(', l) + ";")

# ---------------------------------------------------------
if __name__ == "__main__":
    main()



from default.test_rdm import test_rdm, test_rdm_bw_functional
from efa.efa_common import efa_run_client_server_test

import pytest


@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_pingpong(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong", iteration_type,
                               completion_semantic, memory_type, "all", completion_type=completion_type)

@pytest.mark.functional
@pytest.mark.serial
def test_mr_exhaustion_rdm_pingpong(cmdline_args):
    efa_run_client_server_test(cmdline_args, "fi_efa_exhaust_mr_reg_rdm_pingpong", "short",
                                "transmit_complete", "host_to_host", "all", timeout=1000)

@pytest.mark.functional
def test_rdm_pingpong_range(cmdline_args, completion_semantic, memory_type, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong", "short",
                               completion_semantic, memory_type, message_size)

@pytest.mark.functional
def test_rdm_pingpong_no_inject_range(cmdline_args, completion_semantic, inject_message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong -j 0", "short",
                               completion_semantic, "host_to_host", inject_message_size)

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_pingpong(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_pingpong", iteration_type,
                               completion_semantic, memory_type, "all", completion_type=completion_type)

@pytest.mark.functional
def test_rdm_tagged_pingpong_range(cmdline_args, completion_semantic, memory_type, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_pingpong", "short",
                               completion_semantic, memory_type, message_size)

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_bw(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw", iteration_type,
                               completion_semantic, memory_type, "all", completion_type=completion_type)

@pytest.mark.functional
def test_rdm_tagged_bw_range(cmdline_args, completion_semantic, memory_type, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw", "short",
                               completion_semantic, memory_type, message_size)

@pytest.mark.functional
def test_rdm_tagged_bw_no_inject_range(cmdline_args, completion_semantic, inject_message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw -j 0", "short",
                               completion_semantic, "host_to_host", inject_message_size)

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_atomic(cmdline_args, iteration_type, completion_semantic, memory_type):
    from copy import copy

    from common import ClientServerTest

    if "neuron" in memory_type:
        pytest.skip("Neuron does not fully support atomics")

    # the rdm_atomic test's run time has a high variance when running single c6gn instance.
    # the issue is tracked in:  https://github.com/ofiwg/libfabric/issues/7002
    # to mitigate the issue, set the maximum timeout of fi_rdm_atomic to 1800 seconds.
    cmdline_args_copy = copy(cmdline_args)
    test = ClientServerTest(cmdline_args_copy, "fi_rdm_atomic", iteration_type, completion_semantic,
                            memory_type=memory_type, timeout=1800)
    test.run()

@pytest.mark.functional
def test_rdm_tagged_peek(cmdline_args):
    from copy import copy

    from common import ClientServerTest

    test = ClientServerTest(cmdline_args, "fi_rdm_tagged_peek", timeout=1800)
    test.run()




import os
import sys
import re
import subprocess

class G:
    unit_name = None
    template_lines = []
    tests = []

class RE:
    m = None
    def match(pat, str, flags=0):
        RE.m = re.match(pat, str, flags)
        return RE.m

def main():
    os.chdir(sys.path[0])

    if not os.path.exists("out"):
        os.mkdir("out")

    argc = len(sys.argv)
    if argc > 1:
        args = sys.argv[1:]
    else:
        args = load_test_units()
    for a in args:
        G.unit_name = a
        G.template_lines = []
        G.tests = []
        load_unit(a + '.unit')

        os.chdir("out")
        if build_unit():
            run_tests()
        os.chdir("..")

def load_test_units():
    files = os.listdir()
    units = []
    for f in files:
        if RE.match(r'(\w+)\.unit', f):
            units.append(RE.m.group(1))
    return units

def load_unit(file_spec):
    cur_block = None
    cur_test = None

    with open(file_spec, "r") as In:
        for line in In:
            if RE.match(r'(\w+):', line):
                cur_block = RE.m.group(1)
            elif cur_block == 'template':
                if RE.match(r'\s+\[(\S+):(\w+)\]', line):
                    load_source(RE.m.group(1), RE.m.group(2))
                else:
                    # trim the leading indentation
                    G.template_lines.append(re.sub('^    ', '', line))
            elif cur_block == 'TESTS':
                if RE.match(r'\s+cmd:\s*(.+)', line):
                    cur_test = {'cmd': RE.m.group(1), 'expect': None}
                    G.tests.append(cur_test)
                elif RE.match(r'\s+expect:\s*(.*)', line):
                    cur_test['expect'] = RE.m.group(1).rstrip()

def load_source(file_src, name):
    capture = False
    # assume cwd is test/units
    with open("../../" + file_src, "r") as In:
        for line in In:
            if RE.match(r'\s*\/\*\s*START UNIT\s*-\s*(\w+)', line) and RE.m.group(1) == name:
                capture = True
            elif RE.match(r'\s*\/\*\s*END UNIT', line):
                capture = False
            elif capture:
                G.template_lines.append(line)

def build_unit():
    with open("t.c", "w") as Out:
        for line in G.template_lines:
            Out.write(line)

    if os.path.exists("t"):
        os.remove("t")

    if 'CC' in os.environ:
        cc = os.environ['CC']
    else:
        cc = 'gcc'

    os.system(cc + " -o t t.c")

    return os.path.exists("t")


def run_tests():
    errs = 0
    print("Test - %s" % G.unit_name)
    for i, test in enumerate(G.tests):
        print("  %d:" % (i + 1))
        print("    %s" % test['cmd'])
        ret = subprocess.run(test['cmd'], shell=True, encoding='ascii', capture_output=True)
        if ret.returncode:
            print("    not ok, exit code = %d" % ret.returncode)
            print(ret.stdout)
        else:
            output = ret.stdout.strip()
            print("    %s" % output)
            if test['expect'] and test['expect'] != output:
                print("    not ok, expect: %s" % test['expect'])
                errs += 1
            else:
                print("    ok");
    return errs

#---------------------------------------- 
if __name__ == "__main__":
    main()



##
## Copyright (C) by Argonne National Laboratory
##     See COPYRIGHT in top-level directory
##

from local_python import MPI_API_Global as G
from local_python import RE
import re

def get_kind_map(lang, is_large=False):
    if lang.upper() == 'C':
        if is_large:
            return G.MAPS['BIG_C_KIND_MAP']
        else:
            return G.MAPS['SMALL_C_KIND_MAP']
    elif lang.upper() == 'F08':
        if is_large:
            return G.MAPS['BIG_F08_KIND_MAP']
        else:
            return G.MAPS['SMALL_F08_KIND_MAP']
    elif lang.upper() == 'F90':
        return G.MAPS['SMALL_F90_KIND_MAP']
    elif lang.upper() == "LIS":
        return G.MAPS['LIS_KIND_MAP']
    else:
        raise Exception("Kind Mapping for [%s] not available" % lang)

def function_has_POLY_parameters(func):
    for p in func['parameters']:
        if p['kind'].startswith('POLY'):
            return True
    return False

def get_function_name(func, is_large=False):
    if is_large:
        name = func['name'] + "_c"
        return name
    else:
        name = func['name']
        if 'mpix' in func:
            G.mpix_symbols[name] = "functions"
            name = re.sub(r'MPI_', 'MPIX_', name)
        return name

def split_line_with_break(s, tail, N=100):
    """Breaks a long line with proper indentations.
    This simplistic routine splits on ", ", thus only works with function declarations
    and simple function calls such as those generated by this script. """
    out_list = []

    tlist = []
    n = 0

    # by default, segments indent by additional 4 spaces
    if RE.match(r'(\s*)', s):
        n_lead = len(RE.m.group(1)) + 4

    # -------------
    def break_s_next(s_lead, s_next):
        nonlocal tlist, n
        for a in s_next.split(', '):
            if n == 0:
                # first line
                tlist = [s_lead, a]
                n = n_lead + len(a)
            elif n + 2 + len(a) < N:
                # just append to tlist
                tlist.append(', ')
                tlist.append(a)
                n += 2 + len(a)
            else:
                # break the line
                tlist.append(',')
                out_list.append(''.join(tlist))
                # start new line with leading spaces
                # if lead is too much, it won't look good
                if n_lead > N - 40:
                    tlist = [' ' * 20, a]
                    n = 20 + len(a)
                else:
                    tlist = [' ' * n_lead, a]
                    n = n_lead + len(a)
        # leave last segment with tail

    # -------------
    if len(s) < N:
        tlist.append(s)
        n = len(s)
    elif RE.match(r'(.*?\()(.*)', s):
        # line with function pattern, match indent at opening parenthesis
        s_lead, s_next = RE.m.group(1,2)
        n_lead = len(s_lead)

        break_s_next(s_lead, s_next)
    else:
        break_s_next('', s)

    # tail is mostly for "__attribute__ ((weak, alias(...))));",
    # which contains , that we do not desire to break
    if tail:
        if n + 1 + len(tail) < 100:
            out_list.append(''.join(tlist) + ' ' + tail)
        else:
            out_list.append(''.join(tlist))
            out_list.append(' ' * n_lead + tail)
    else:
        out_list.append(''.join(tlist))

    return out_list

def get_C_param(param, func, mapping):
    kind = param['kind']
    if kind == "VARARGS":
        return "..."

    want_star, want_bracket = '', ''
    param_type = mapping[kind]

    if param['func_type']:
        param_type = param['func_type']
        if mapping['_name'].startswith("BIG_"):
            param_type += "_c"

    if param_type in G.mpix_symbols:
        param_type = re.sub(r'MPI_', 'MPIX_', param_type)

    if not param_type:
        raise Exception("Type mapping [%s] %s not found!" % (mapping, kind))

    # We special treat a few cases to simplify the general rules
    if kind == "ARGUMENT_COUNT":
        if re.match(r'mpi_info_create_env', func['name'], re.IGNORECASE):
            # -> int argc
            pass
        else:
            # MPI_Init, MPI_Init_thread -> int *argc
            want_star = 1
    elif kind == "ARGUMENT_LIST":
        if re.match(r'mpi_info_create_env', func['name'], re.IGNORECASE):
            # -> char *argv[]
            want_star = 1
            want_bracket = 1
        else:
            # MPI_Init, MPI_Init_thread -> char ***argv
            want_star = 3
    elif kind == "STATUS" and param['length'] is not None:
        # MPI_{Wait,Test}{all,some}
        # gcc-11 warns when we pass MPI_STATUSES_IGNORE to MPI_Status statues[]
        want_star = 1
    elif not want_star:
        if is_pointer_type(param):
            if kind == "STRING_ARRAY":
                want_star = 1
                want_bracket = 1
            elif kind == "STRING_2DARRAY":
                want_star = 2
                want_bracket = 1
            elif param['pointer'] is not None and not param['pointer']:
                want_bracket = 1
            elif param['length'] is not None and kind != "STRING":
                want_bracket = 1
            else:
                want_star = 1

    s = ''
    if param['constant']:
        s += "const "
    s += param_type

    if want_star:
        s += " " + "*" * want_star
    else:
        s += " "
    s += param['name']

    if want_bracket:
        s += "[]"
    if isinstance(param['length'], list):
        s += "[%s]" % param['length'][-1]

    return s

def is_pointer_type(param):
    if RE.match(r'(STRING\w*)$', param['kind']):
        return 1
    elif RE.match(r'(STATUS|F90_STATUS|F08_STATUS)$', param['kind']):
        return 1
    elif RE.match(r'(ATTRIBUTE_VAL\w*|(C_)?BUFFER\d?|EXTRA_STATE\d*|TOOL_MPI_OBJ|(POLY)?FUNCTION\w*)$', param['kind']):
        return 1
    elif param['param_direction'] != 'in':
        return 1
    elif param['length']:
        return 1
    elif param['pointer']:
        return 1
    else:
        return 0

def param_is_in(p):
    if p['param_direction'] == 'in' or p['param_direction'] == 'inout':
        return True
    else:
        return False

def param_is_out(p):
    if p['param_direction'] == 'out' or p['param_direction'] == 'inout':
        return True
    else:
        return False

def get_userbuffer_group(func_name, parameters, i):
    """internal function used by process_func_parameters"""
    p = parameters[i]
    p2 = parameters[i + 1]
    p3 = parameters[i + 2]
    if RE.match(r'mpi_i?(alltoall|allgather|gather|scatter)', func_name, re.IGNORECASE):
        type = "inplace"
        if RE.search(r'send', p['name'], re.IGNORECASE) and RE.search(r'scatter', func_name, re.IGNORECASE):
            type = "noinplace"
        elif RE.search(r'recv', p['name'], re.IGNORECASE) and not RE.search(r'scatter', func_name, re.IGNORECASE):
            type = "noinplace"

        if RE.search(r'alltoallw', func_name, re.IGNORECASE):
            group_kind = "USERBUFFER-%s-w" % (type)
            group_count = 4
        elif p3['kind'] == "DATATYPE":
            group_kind = "USERBUFFER-%s" % (type)
            group_count = 3
        else:
            group_kind = "USERBUFFER-%s-v" % (type)
            group_count = 4
    elif RE.match(r'mpi_i?neighbor', func_name, re.IGNORECASE):
        if RE.search(r'alltoallw', func_name, re.IGNORECASE):
            group_kind = "USERBUFFER-neighbor-w"
            group_count = 4
        elif p3['kind'] == "DATATYPE":
            group_kind = "USERBUFFER-neighbor"
            group_count = 3
        else:
            group_kind = "USERBUFFER-neighbor-v"
            group_count = 4
    elif RE.match(r'mpi_i?(allreduce|reduce|scan|exscan)', func_name, re.IGNORECASE):
        group_kind = "USERBUFFER-reduce"
        group_count = 5
    elif RE.match(r'mpi_p(send|recv)_init', func_name, re.IGNORECASE):
        group_kind = "USERBUFFER-partition"
        group_count = 4
    elif RE.search(r'XFER_NUM_ELEM', p2['kind']) and RE.search(r'DATATYPE', p3['kind']):
        group_kind = "USERBUFFER-simple"
        group_count = 3
    else:
        group_kind, group_count = None, 0
    return (group_kind, group_count)
